@@ -1,10 +1,17 @@
-use dioxus::{nodes::VSuspended, prelude::*, DomEdit, TestDom};
+#![allow(non_snake_case)]
+
+use dioxus::prelude::*;
 use dioxus_core as dioxus;
 use dioxus_core_macro::*;
 use dioxus_html as dioxus_elements;
 
-static Parent: FC<()> = |cx, props| {
-    let value = cx.use_hook(|_| String::new(), |f| &*f, |_| {});
+#[test]
+fn test_borrowed_state() {
+    let _ = VirtualDom::new(Parent);
+}
+
+fn Parent(cx: Scope<()>) -> Element {
+    let value = cx.use_hook(|_| String::new(), |f| &*f);
 
     cx.render(rsx! {
         div {
@@ -14,32 +21,29 @@ static Parent: FC<()> = |cx, props| {
             Child { name: value }
         }
     })
-};
+}
 
 #[derive(Props)]
 struct ChildProps<'a> {
-    name: &'a String,
+    name: &'a str,
 }
 
-fn Child<'a>(cx: Context<'a>, props: &'a ChildProps) -> DomTree<'a> {
+fn Child<'a>(cx: Scope<'a, ChildProps<'a>>) -> Element {
     cx.render(rsx! {
         div {
             h1 { "it's nested" }
-            Child2 { name: props.name }
+            Child2 { name: cx.props.name }
         }
     })
 }
 
 #[derive(Props)]
 struct Grandchild<'a> {
-    name: &'a String,
+    name: &'a str,
 }
 
-fn Child2<'a>(cx: Context<'a>, props: &Grandchild) -> DomTree<'a> {
+fn Child2<'a>(cx: Scope<'a, Grandchild<'a>>) -> Element {
     cx.render(rsx! {
-        div { "Hello {props.name}!" }
+        div { "Hello {cx.props.name}!" }
     })
 }
-
-#[test]
-fn test_borrowed_state() {}
